
#region Header

/* /////////////////////////////////////////////////////////////////////////////////////////////////
PLUGIN for Virt-A-Mate by Fabio Den.

Written in Notepad++
	
Version 1.0 2021-03-30
	Initial release.

License: CC BY-SA
	
///////////////////////////////////////////////////////////////////////////////////////////////// */

using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using UnityEngine.UI;
using UnityEngine.Events;
using UnityEngine.Animations;
using System.Linq;
using MVR.FileManagementSecure;
using SimpleJSON;

#endregion

public class Bookmarks : MVRScript
{

	#region Variables

	private VRWebBrowser browser = null;
	private PopupEx popup;
	private bool isDeleting = false;
	private bool isDeletingCat = false;
	private float sep = 0.0F;
	private LabelEx message;
	private ButtonEx yesbtn;
	private ButtonEx nobtn;
	private SpacerEx sp1;
	private SpacerEx sp2;
	private int currentCateID = 0;
	private int currentBookID = 0;
	List<Category> cates = new List<Category>();
	private int Counter = 0;
	private LabelEx label;
	private ButtonEx okbtn;
	private const string _saveExt = "json";
	private const string _saveFolder = "Custom/Assets/fabio/Bookmarks/";

	#endregion

	#region Events

	public override void Init()
	{
		Install();
		SetupGUI();
	}

	public void Start()
	{
		ShowConfirmDel(false);
		ShowConfirmEdit(false);
		LoadData();

		if (cates.Count > 0)
        {
			Category cat = cates[0];

			if (cat != null)
			{
				FilterCategory(cat.id);
			}
		}
	}

	protected void OnSceneLoaded()
	{
		foreach (Atom a in SuperController.singleton.GetAtoms())
		{
			if ((a.type == "WebPanelEmissive") || (a.type == "WebPanel") || (a.type == "WebBrowser"))
            {
				SyncPopup(a.name);
				break;
            }
				
		}
	}

	public void OnEnable()
	{
		SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;
	}

	public void OnDisable()
	{
		SuperController.singleton.onSceneLoadedHandlers -= OnSceneLoaded;
	}

	public void OnDestroy()
	{
		SaveData();

		foreach (Category cat in cates)
		{
			foreach (Bookmark book in cat.books)
			{
				book.Remove();
			}

			cat.books.Clear();
			cat.Remove();
		}

		cates.Clear();
	}

	#endregion

	#region GUI

	private void SetupGUI()
	{
		// header
		GUICreateSpacer(this, false, sep);
		GUICreateSpacer(this, true, sep);
		popup = GUICreatePopup(this, "mypopup", "Select Browser", false, SyncPopupChoices, SyncPopup);
		ButtonEx btn = GUICreateButton(this, "Save Bookmarks", new Color(0.78F, 0.78F, 0.899F), true, true, SaveData);
		btn.button.height = 118.0F;
		BClr.Paint(btn, BClr.green, BClr.white);
		btn = GUICreateButton(this, "Back", new Color(0.78F, 0.78F, 0.899F), true, false, GoBack);
		btn.button.height = 60F;
		BClr.Paint(btn, BClr.purple, BClr.white);
		btn = GUICreateButton(this, "Forward", new Color(0.78F, 0.78F, 0.899F), true, true, GoForward);
		btn.button.height = 60F;
		BClr.Paint(btn, BClr.purple, BClr.white);
		btn = GUICreateButton(this, "Home", new Color(0.78F, 0.78F, 0.899F), true, false, GoHome);
		btn.button.height = 60F;
		BClr.Paint(btn, BClr.purple, BClr.white);
		btn = GUICreateButton(this, "Paste from Clipboard", new Color(0.78F, 0.78F, 0.899F), true, true, Paste);
		btn.button.height = 60F;
		BClr.Paint(btn, BClr.purple, BClr.white);

		// manager
		//GUICreateSpacer(this, false, sep);
		//GUICreateSpacer(this, true, sep);
		btn = GUICreateButton(this, "Add Category", new Color(0.78F, 0.78F, 0.899F), true, false, AddCategory);
		btn.button.height = 60F;
		BClr.Paint(btn, BClr.blue, BClr.white);
		btn = GUICreateButton(this, "Add Bookmark", new Color(0.78F, 0.78F, 0.899F), true, true, AddBookmark);
		btn.button.height = 60F;
		BClr.Paint(btn, BClr.green, BClr.white);
		btn = GUICreateButton(this, "Rename", new Color(0.78F, 0.78F, 0.899F), true, false, RenameCategory);
		btn.button.height = 60F;
		BClr.Paint(btn, BClr.purple, BClr.white);
		btn = GUICreateButton(this, "Rename", new Color(0.78F, 0.78F, 0.899F), true, true, RenameBookmark);
		btn.button.height = 60F;
		BClr.Paint(btn, BClr.purple, BClr.white);
		btn = GUICreateButton(this, "Delete", new Color(0.78F, 0.78F, 0.899F), true, false, DeleteCategory);
		btn.button.height = 60F;
		BClr.Paint(btn, BClr.purple, BClr.white);
		btn = GUICreateButton(this, "Delete", new Color(0.78F, 0.78F, 0.899F), true, true, DeleteBookmark);
		btn.button.height = 60F;
		BClr.Paint(btn, BClr.purple, BClr.white);

		// confirm
		sp1 = GUICreateSpacer(this, false, sep);
		sp2 = GUICreateSpacer(this, true, sep);
		message = GUICreateLabel(this, " ", 116.0F, false);
		yesbtn = GUICreateButton(this, "Yes", new Color(0.78F, 0.78F, 0.899F), true, true, ConfirmYes);
		BClr.Paint(yesbtn, BClr.ngreen, BClr.white);
		nobtn = GUICreateButton(this, "No", new Color(0.78F, 0.78F, 0.899F), true, true, ConfirmNo);
		BClr.Paint(nobtn, BClr.red, BClr.white);
		label = CreateStringTextbox(true);
		okbtn = GUICreateButton(this, "OK", new Color(0.78F, 0.78F, 0.899F), true, true, ConfirmOK);
		BClr.Paint(okbtn, BClr.ngreen, BClr.white);

		GUICreateSpacer(this, false, sep);
		GUICreateSpacer(this, true, sep);

		// content
		// left categories
		// right bookmarks
	}

	#endregion

	#region Header

	private void SelectAtom()
	{
		if (isDeleting == false)
		{
			Atom atom = SuperController.singleton.GetAtomByUid(popup.value.val);

			if (atom != null)
			{
				SelectAtom(atom);
			}
		}
	}

	public static void SelectAtom(Atom atom)
	{
		if (atom != null)
		{
			FreeControllerV3 ctrl = atom.GetStorableByID("control") as FreeControllerV3;

			if (ctrl != null)
				SuperController.singleton.SelectController(ctrl);
		}
	}

	// before opening the popup
	private void SyncPopupChoices()
	{
		// example, fill with people
		List<string> choices = new List<string>();

		foreach (Atom a in SuperController.singleton.GetAtoms())
		{
			if ((a.type == "WebPanelEmissive") || (a.type == "WebPanel") || (a.type == "WebBrowser"))
				choices.Add(a.name);
		}

		popup.value.choices = choices;
	}

	// after choosing a string from the popup
	private void SyncPopup(string atomUID)
	{
		popup.value.val = atomUID;

		Atom atom = SuperController.singleton.GetAtomByUid(atomUID);

		if (atom != null)
		{
			browser = atom.GetStorableByID("BrowserGUI") as VRWebBrowser;
		}
		else
		{
			browser = null;
		}
	}

	private void GoBack()
	{
		if (isDeleting == false)
		{
			BrowserBack();
		}
	}

	private void GoForward()
	{
		if (isDeleting == false)
		{
			BrowserForward();
		}
	}

	private void GoHome()
	{
		if (isDeleting == false)
		{
			BrowserHome();
		}
	}

	private void Paste()
	{
		if (isDeleting == false)
		{
			BrowserPaste();
		}
	}

	#endregion

	#region Confirm

	private void ConfirmYes()
	{
		if (isDeleting == true)
		{
			if (isDeletingCat)
            {
				// delete current Category
				Category cat = getCategory(currentCateID);

				if (cat != null)
				{
					foreach (Bookmark book in cat.books)
					{
						book.Remove();
					}

					cat.books.Clear();
					cat.Remove();
					cates.Remove(cat);
					currentBookID = 0;
					currentCateID = 0;
				}
			}
			else
            {
				// delete current Bookmark
				Category cat = getCategory(currentCateID);

				if (cat != null)
                {
					Bookmark _book = getBookmark(cat, currentBookID);

					if (_book != null)
					{
						cat.books.Remove(_book);
						_book.Remove();
						currentBookID = 0;
						cat.SetCurrent(true);
					}
				}
			}

			ShowConfirmDel(false);
			isDeleting = false;
		}
	}

	private void ConfirmNo()
	{
		if (isDeleting == true)
		{
			ShowConfirmDel(false);
			isDeleting = false;
		}
	}

	private void ShowConfirmDel(bool visible)
	{
		if (visible)
		{
			sp1.Show();
			sp2.Show();
			message.Show();
			yesbtn.Show();
			nobtn.Show();
		}
		else
		{
			sp1.Hide();
			sp2.Hide();
			message.Hide();
			yesbtn.Hide();
			nobtn.Hide();
		}
	}

	private void ShowConfirmEdit(bool visible)
	{
		if (visible)
		{
			sp1.Show();
			sp2.Show();
			message.Show();
			label.Show();
			okbtn.Show();
		}
		else
		{
			sp1.Hide();
			sp2.Hide();
			message.Hide();
			label.Hide();
			okbtn.Hide();
		}
	}

	private void ConfirmOK()
	{
		if (isDeleting == true)
		{
			if (isDeletingCat)
			{
				// edit current Category
				Category cat = getCategory(currentCateID);

				if (cat != null)
				{
					if (label.value.val.Length > 0)
                    {
						cat.name = label.value.val;
						cat.button.button.label = cat.name;
                    }
				}
			}
			else
			{
				// delete current Bookmark
				Category cat = getCategory(currentCateID);

				if (cat != null)
				{
					Bookmark _book = getBookmark(cat, currentBookID);

					if (_book != null)
					{
						if (label.value.val.Length > 0)
						{
							_book.title = label.value.val;
							_book.SetCurrent(true);
						}
					}
				}
			}

			ShowConfirmEdit(false);
			isDeleting = false;
		}
	}

	#endregion

	#region Browser

	private string BrowserGetURL()
    {
		return browser.url;
    }

	private string BrowserGetTitle()
	{
		return browser.navigatedURL;
	}

	private void BrowserNavigate(string url)
    {
		if (browser != null)
		{
			browser.url = url;
		}
	}

	private void BrowserHome()
	{
		if (browser != null)
		{
			browser.url = "https://www.google.com";
		}
	}

	private void BrowserBack()
	{
		if (browser != null)
		{
			browser.GoBack();
		}
	}

	private void BrowserForward()
	{
		if (browser != null)
		{
			browser.GoForward();
		}
	}

	private void BrowserPaste()
	{
		if (browser != null)
		{
			browser.CopyURLFromClipboard();
		}
	}

	#endregion

	#region Control

	public class LabelEx
	{
		public UIDynamicTextField label;
		public Transform _transform;
		public JSONStorableString value;

		// Constructor (not necessary, but helpful)
		public LabelEx(UIDynamicTextField label, Transform _transform)
		{
			this.label = label;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public LabelEx GUICreateLabel(MVRScript dll, string label, float height, bool rightside)
	{
		JSONStorableString value = new JSONStorableString("", "");
		LabelEx gui = CreateTextField2(value, rightside);
		gui.label.text = label;
		gui.label.height = height;
		gui.value = value;

		return gui;
	}

	public LabelEx CreateTextField2(JSONStorableString jss, bool rightSide = false)
	{
		LabelEx label = null;

		UIDynamicTextField uIDynamicTextField = null;
		if (manager != null && manager.configurableTextFieldPrefab != null && jss.text == null)
		{
			Transform transform = CreateUIElement(manager.configurableTextFieldPrefab.transform, rightSide);
			if (transform != null)
			{
				uIDynamicTextField = transform.GetComponent<UIDynamicTextField>();
				if (uIDynamicTextField != null)
				{
					textFieldToJSONStorableString.Add(uIDynamicTextField, jss);
					jss.dynamicText = uIDynamicTextField;

					label = new LabelEx(uIDynamicTextField, transform);
				}
			}
		}

		return label;
	}

	public class SpacerEx
	{
		public UIDynamic spacer;
		public Transform _transform;

		// Constructor (not necessary, but helpful)
		public SpacerEx(UIDynamic spacer, Transform _transform)
		{
			this.spacer = spacer;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public SpacerEx GUICreateSpacer(MVRScript dll, bool rightside, float height)
	{
		SpacerEx gui = CreateSpacer2(rightside);
		gui.spacer.height = height;

		return gui;
	}

	public SpacerEx CreateSpacer2(bool rightSide = false)
	{
		SpacerEx spacer = null;

		UIDynamic result = null;
		if (manager != null && manager.configurableSpacerPrefab != null)
		{
			Transform transform = CreateUIElement(manager.configurableSpacerPrefab.transform, rightSide);
			if (transform != null)
			{
				result = transform.GetComponent<UIDynamic>();

				spacer = new SpacerEx(result, transform);
			}
		}
		return spacer;
	}

	public class ButtonEx
	{
		public UIDynamicButton button;
		public Transform _transform;

		// Constructor (not necessary, but helpful)
		public ButtonEx(UIDynamicButton button, Transform _transform)
		{
			this.button = button;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public ButtonEx GUICreateButton(MVRScript dll, string label, Color _color, bool useColor, bool rightside, UnityAction callback)
	{
		ButtonEx gui = CreateButton2(label, rightside);
		if (useColor) { gui.button.buttonColor = _color; }
		gui.button.button.onClick.AddListener(callback);

		return gui;
	}

	public ButtonEx CreateButton2(string label, bool rightSide = false)
	{
		ButtonEx button = null;

		UIDynamicButton uIDynamicButton = null;
		if (manager != null && manager.configurableButtonPrefab != null)
		{
			Transform transform = CreateUIElement(manager.configurableButtonPrefab.transform, rightSide);
			if (transform != null)
			{
				uIDynamicButton = transform.GetComponent<UIDynamicButton>();
				if (uIDynamicButton != null)
				{
					uIDynamicButton.label = label;

					button = new ButtonEx(uIDynamicButton, transform);
				}
			}
		}

		return button;
	}

	public class ToggleEx
	{
		public UIDynamicToggle toggle;
		public JSONStorableBool value;
		public Transform _transform;

		// Constructor (not necessary, but helpful)
		public ToggleEx(UIDynamicToggle toggle, Transform _transform)
		{
			this.toggle = toggle;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public ToggleEx GUICreateCheckbox(MVRScript dll, string name, string label, bool startvalue, bool rightside, JSONStorableBool.SetJSONBoolCallback callback)
	{
		JSONStorableBool gui = new JSONStorableBool(name, startvalue, callback);
		ToggleEx Toggle = CreateToggle2(gui, rightside);
		Toggle.toggle.label = label;
		dll.RegisterBool(gui);
		Toggle.value = gui;

		return Toggle;
	}

	public ToggleEx CreateToggle2(JSONStorableBool jsb, bool rightSide = false)
	{
		ToggleEx toggle = null;

		UIDynamicToggle uIDynamicToggle = null;
		if (manager != null && manager.configurableTogglePrefab != null && jsb.toggle == null)
		{
			Transform transform = CreateUIElement(manager.configurableTogglePrefab.transform, rightSide);
			if (transform != null)
			{
				uIDynamicToggle = transform.GetComponent<UIDynamicToggle>();
				if (uIDynamicToggle != null)
				{
					toggleToJSONStorableBool.Add(uIDynamicToggle, jsb);
					uIDynamicToggle.label = jsb.name;
					jsb.toggle = uIDynamicToggle.toggle;

					toggle = new ToggleEx(uIDynamicToggle, transform);
				}
			}
		}

		return toggle;
	}

	public class SliderEx
	{
		public UIDynamicSlider slider;
		public JSONStorableFloat value;
		public Transform _transform;

		// Constructor (not necessary, but helpful)
		public SliderEx(UIDynamicSlider slider, Transform _transform)
		{
			this.slider = slider;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public SliderEx GUICreateSlider(MVRScript dll, string name, string label, float startvalue, bool rightside, float minvalue, float maxvalue, UnityAction<float> _event)
	{
		JSONStorableFloat gui = new JSONStorableFloat(name, startvalue, minvalue, maxvalue);
		SliderEx Slider = CreateSlider2(gui, rightside);
		Slider.slider.label = label;
		Slider.value = gui;
		dll.RegisterFloat(gui);
		Slider.value = gui;
		Slider.slider.slider.onValueChanged.AddListener(_event);

		return Slider;
	}

	public SliderEx CreateSlider2(JSONStorableFloat jsf, bool rightSide = false)
	{
		SliderEx slider = null;

		UIDynamicSlider uIDynamicSlider = null;
		if (manager != null && manager.configurableSliderPrefab != null && jsf.slider == null)
		{
			Transform transform = CreateUIElement(manager.configurableSliderPrefab.transform, rightSide);
			if (transform != null)
			{
				uIDynamicSlider = transform.GetComponent<UIDynamicSlider>();
				if (uIDynamicSlider != null)
				{
					uIDynamicSlider.Configure(jsf.name, jsf.min, jsf.max, jsf.val, jsf.constrained, "F2", true, !jsf.constrained);
					jsf.slider = uIDynamicSlider.slider;
					sliderToJSONStorableFloat.Add(uIDynamicSlider, jsf);

					slider = new SliderEx(uIDynamicSlider, transform);
				}
			}
		}
		return slider;
	}

	public class PopupEx
	{
		public UIDynamicPopup popup;
		public JSONStorableStringChooser value;
		public Transform _transform;

		// Constructor (not necessary, but helpful)
		public PopupEx(UIDynamicPopup popup, Transform _transform)
		{
			this.popup = popup;
			this._transform = _transform;
		}

		public void Show()
		{
			_transform.gameObject.SetActive(value: true);
		}

		public void Hide()
		{
			_transform.gameObject.SetActive(value: false);
		}
	}

	public PopupEx GUICreatePopup(MVRScript dll, string name, string label, bool rightside, UIPopup.OnOpenPopup popup, JSONStorableStringChooser.SetStringCallback callback)
	{
		JSONStorableStringChooser gui = new JSONStorableStringChooser(name, null, "", "", callback);
		dll.RegisterStringChooser(gui);
		PopupEx dp = CreateFilterablePopup2(gui, rightside);
		dp.popup.popupPanelHeight = 500.0F;
		dp.popup.label = label;
		dp.popup.popup.onOpenPopupHandlers = popup;
		dp.value = gui;
		return dp;
	}

	public PopupEx CreateFilterablePopup2(JSONStorableStringChooser jsc, bool rightSide = false)
	{
		PopupEx popup = null;

		UIDynamicPopup uIDynamicPopup = null;
		if (manager != null && manager.configurableFilterablePopupPrefab != null && jsc.popup == null)
		{
			Transform transform = CreateUIElement(manager.configurableFilterablePopupPrefab.transform, rightSide);
			if (transform != null)
			{
				uIDynamicPopup = transform.GetComponent<UIDynamicPopup>();
				if (uIDynamicPopup != null)
				{
					popupToJSONStorableStringChooser.Add(uIDynamicPopup, jsc);
					uIDynamicPopup.label = jsc.name;
					jsc.popup = uIDynamicPopup.popup;

					popup = new PopupEx(uIDynamicPopup, transform);
				}
			}
		}

		return popup;
	}

	public LabelEx CreateStringTextbox(bool rhs)
	{
		JSONStorableString json = new JSONStorableString("", "");
		LabelEx textField = CreateTextField2(json, rhs);
		SetTextFieldHeight(textField.label, 60);
		InputField input = textField.label.gameObject.AddComponent<InputField>();
		input.textComponent = textField.label.UItext;
		input.textComponent.fontSize = 32; // 32
		textField.label.backgroundColor = Color.white;
		json.inputField = input;
		textField.value = json;
		input.onValueChanged.AddListener((string value) => {
			json.val = value;
		});
		return textField;
	}

	public void SetTextFieldHeight(UIDynamicTextField textField, int height)
	{
		LayoutElement component = textField.GetComponent<LayoutElement>();
		if (component != null)
		{
			component.minHeight = height;
			component.preferredHeight = height;
		}
		textField.height = height;
	}

	#endregion

	#region Color

	public class BClr
	{
		public static Color green = new Color32(70, 149, 149, 255);
		public static Color blue = new Color32(43, 130, 193, 255);
		public static Color purple = new Color32(150, 85, 159, 255);
		public static Color white = new Color32(255, 255, 255, 255);
		public static Color red = new Color32(152, 12, 99, 255);
		public static Color ngreen = new Color32(40, 95, 95, 255);

		public static void Paint(ButtonEx btn, Color back, Color text)
		{
			btn.button.buttonColor = back;
			btn.button.textColor = text;
		}
	}

	#endregion

	#region Category

	public class Category
	{
		private Bookmarks parent;
		public ButtonEx button;
		public int id;
		public string name;
		public List<Bookmark> books = new List<Bookmark>();

		public void SetCurrent(bool current)
		{
			if (current)
            {
				char uni = '\u2713';
				BClr.Paint(button, BClr.blue, BClr.white);
				button.button.label = uni.ToString() + " <b>" + name + "</b> ";
				button.button.label += "\n" + books.Count.ToString() + " websites";
			}
			else
            {
				BClr.Paint(button, BClr.blue, BClr.white);
				button.button.label = "<b>" + name + "</b>";
				button.button.label += "\n" + books.Count.ToString() + " websites";
			}

			parent.currentCateID = id;
		}

		public Category(Bookmarks parent, string name)
		{
			this.parent = parent;
			this.name = name;
			parent.Counter++;
			this.id = parent.Counter;
			CreateCategory();
		}

		private void CreateCategory()
		{
			button = parent.GUICreateButton(parent, name, new Color(0.78F, 0.78F, 0.899F), true, false, SelectCategory);
			button.button.height = 100F;
		}

		public void SelectCategory()
		{
			// set as current category
			if (parent.isDeleting == false)
			{
				parent.FilterCategory(id);
			}
		}

		public void Remove()
		{
			foreach (Bookmark book in books)
			{
				book.Remove();
			}

			books.Clear();

			parent.RemoveButton(button.button);
		}
	}

	private void FilterCategory(int id)
    {
		if (isDeleting == false)
		{
			foreach (Category cat in cates)
			{
				if (cat.id == id)
				{
					cat.SetCurrent(true);

					foreach (Bookmark book in cat.books)
					{
						book.SetVisible(true);
					}
				}
				else
				{
					cat.SetCurrent(false);

					foreach (Bookmark book in cat.books)
					{
						book.SetVisible(false);
					}
				}
			}

			currentCateID = id;
		}
	}

	private Category CreateCategory(string name, bool select)
    {
		if (name.Length == 0) { name = "Category#" + (cates.Count + 1).ToString(); }
		Category cat = new Category(this, name);
		cates.Add(cat);
		if (select) { cat.SelectCategory(); }

		return cat;
	}

	private void AddCategory()
	{
		if (isDeleting == false)
		{
			CreateCategory("", true);
		}
	}

	private void RenameCategory()
	{
		if (isDeleting == false)
		{
			Category cat = getCategory(currentCateID);

			if (cat != null)
			{
				string name = cat.name;
				isDeleting = true;
				isDeletingCat = true;
				label.value.val = name;
				message.value.val = "Please enter the new name for this category";
				ShowConfirmEdit(true);
			}
		}
	}

	private void DeleteCategory()
	{
		if (isDeleting == false)
		{
			Category cat = getCategory(currentCateID);

			if (cat != null)
			{
				string name = cat.name;
				isDeleting = true;
				isDeletingCat = true;
				message.value.val = "This action cannot be undone. Are you sure you want to delete <b>" + name + "</b> category?";
				ShowConfirmDel(true);
			}
		}
	}

	private Category getCategory(int id)
    {
		Category _cat = null;

		// iterate categories
		foreach (Category cat in cates)
		{
			if (cat.id == id)
            {
				_cat = cat;
				break;
            }
		}

		return _cat;
    }

	#endregion

	#region Bookmark

	public static string Left(string param, int length)
	{
		string result = param.Substring(0, length);
		return result;
	}

	public class Bookmark
	{
		private Bookmarks parent;
		public ButtonEx button;
		public int id;
		public string title;
		public string url;
		private Category categ;

		public void SetCurrent(bool current)
		{
			string _url = url;
			if (_url.Contains("https://")) { _url = _url.Replace("https://", ""); }
			if (_url.Contains("www.")) { _url = _url.Replace("www.", ""); }
			if (_url.Length > 32) { _url = Left(_url, 28) + "..."; }	

			if (current)
            {
				char uni = '\u25B6';
				button.button.label = uni.ToString() + " <b>" + title + "</b> ";
				button.button.label += "\n" + _url;
				BClr.Paint(button, BClr.green, BClr.white);
			}
			else
            {
				button.button.label = "<b>" + title + "</b>";
				button.button.label += "\n" + _url;
				BClr.Paint(button, BClr.green, BClr.white);
			}

			parent.currentBookID = id;
		}

		public void SetVisible(bool visible)
		{
			if (visible)
				button.Show();
			else
				button.Hide();
		}

		public Bookmark(Bookmarks parent, string title, string url, Category categ)
		{
			this.parent = parent;
			this.title = title;
			this.url = url;
			parent.Counter++;
			this.id = parent.Counter;
			this.categ = categ;
			CreateBookmark();
		}

		private void CreateBookmark()
		{
			button = parent.GUICreateButton(parent, title, new Color(0.78F, 0.78F, 0.899F), true, true, SelectBookmark);
			button.button.height = 100F;
		}

		public void SelectBookmark()
		{
			if (parent.isDeleting == false)
			{
				foreach (Bookmark book in categ.books)
				{
					book.SetCurrent(false);
				}

				SetCurrent(true);
				parent.BrowserNavigate(url);
			}
		}

		public void Remove()
		{
			parent.RemoveButton(button.button);
		}
	}

	private void CreateBookmark(Category cat, string title, string url, bool select)
    {
		if (title.Length == 0) { title = "Untitled#" + (cat.books.Count + 1).ToString(); }
		Bookmark book = new Bookmark(this, title, url, cat);
		cat.books.Add(book);
		book.SetCurrent(false);
		if (select)
		{ 
			book.SelectBookmark();
			cat.SetCurrent(true);
		}
	}

	private void AddBookmark()
	{
		if (isDeleting == false)
		{
			Category cat = getCategory(currentCateID);

			if (cat != null)
			{
				string url = BrowserGetURL();

				if (url.Length > 0)
                {
					string title = BrowserGetTitle();
					CreateBookmark(cat, title, url, true);
				}
			}
		}
	}

	private void RenameBookmark()
	{
		if (isDeleting == false)
		{
			Category cat = getCategory(currentCateID);

			if (cat != null)
			{
				Bookmark _book = getBookmark(cat, currentBookID);

				if (_book != null)
				{
					string name = _book.title;
					isDeleting = true;
					isDeletingCat = false;
					label.value.val = name;
					message.value.val = "Please enter the new name for this bookmark";
					ShowConfirmEdit(true);
				}
			}
		}
	}

	private void DeleteBookmark()
	{
		if (isDeleting == false)
		{
			Category cat = getCategory(currentCateID);

			if (cat != null)
			{
				Bookmark _book = getBookmark(cat, currentBookID);

				if (_book != null)
				{
					string name = _book.title;
					isDeleting = true;
					isDeletingCat = false;
					message.value.val = "This action cannot be undone. Are you sure you want to delete <b>" + name + "</b> bookmark?";
					ShowConfirmDel(true);
				}
			}
		}
	}

	private Bookmark getBookmark(Category cat, int id)
	{
		Bookmark _book = null;

		// iterate categories
		foreach (Bookmark book in cat.books)
		{
			if (book.id == id)
			{
				_book = book;
				break;
			}
		}

		return _book;
	}

	#endregion

	#region Default

	public void DefaultData()
	{
		Category cat = CreateCategory("Music", false);
		CreateBookmark(cat, "Rock", "https://www.youtube.com/watch?v=8SbUC-UaAxE&list=PLJ7kq_XdaWUDTTe4Acw2jXoQ7cdKrg7rb", false);
		CreateBookmark(cat, "Cyberpunk", "https://www.youtube.com/watch?v=wcOhAodaW5Y&list=PLcx8DGkalsx4JPyF7VmDIOg1rzLR-suB4", false);
		CreateBookmark(cat, "Electronic", "https://www.youtube.com/watch?v=kdemFfbS5H0&list=PLG3-q8B8DlBLVW8L604yGeFnCLnFLbSzS", false);
		CreateBookmark(cat, "Remixes", "https://www.youtube.com/watch?v=FXqp9WiFWzc&list=PLe94KlwIw7pLQJkGRnjcaVMCXrJIvxDvC", false);
		CreateBookmark(cat, "Heavy Metal", "https://www.youtube.com/watch?v=eVTXPUF4Oz4&list=PLWUXmX2htJ7Nx2_luG8WAVEnShOW-lGal", false);
		CreateBookmark(cat, "Dark-Ambient", "https://www.youtube.com/watch?v=iXFzstSDx2U&list=PLZUwBQRNk8PDemdD1nxK-r0T8BpfEsnrk", false);
		CreateBookmark(cat, "Fantasy", "https://www.youtube.com/watch?v=S_EJ14KkGks&list=PLOoRHiSvRhgKk1kRAzUk2YseOvoVj0HXE", false);
		CreateBookmark(cat, "Action Tracks", "https://www.youtube.com/watch?v=wPIzcu6jZPM&list=PL2WYvHoUVFEIwfvF7ZhBfcSaAq6f1g5iu", false);
		CreateBookmark(cat, "Casual Songs", "https://www.youtube.com/watch?v=oVcG9lpZV24&list=PLwf-0NjVCR1nZ0idDz_VicqY8JzkWyGUh", false);
		CreateBookmark(cat, "Soft Tracks", "https://www.youtube.com/watch?v=9YgmMJJ34k4&list=PLxZVygmS5cw7Hbi9WmjkyJdeJTzaswzVs", false);

		cat = CreateCategory("Virt-A-Mate", false);
		CreateBookmark(cat, "What's new", "https://hub.virtamate.com/whats-new/", false);
		CreateBookmark(cat, "Resources", "https://hub.virtamate.com/resources/", false);
		CreateBookmark(cat, "Trending resources", "https://hub.virtamate.com/trending/resources.2/", false);
		CreateBookmark(cat, "Forum", "https://hub.virtamate.com/forums/", false);
		CreateBookmark(cat, "Trending threads", "https://hub.virtamate.com/trending/threads.1/", false);
		CreateBookmark(cat, "New posts", "https://hub.virtamate.com/whats-new/posts/416427/", false);
		CreateBookmark(cat, "Media", "https://hub.virtamate.com/media/", false);
		CreateBookmark(cat, "Announcements", "https://hub.virtamate.com/forums/announcements.11/", false);

		cat = CreateCategory("Adult Sites", false);
		CreateBookmark(cat, "XVideos", "https://www.xvideos.com/", false);
		CreateBookmark(cat, "Spankbang", "https://spankbang.com", false);
		CreateBookmark(cat, "Redtube", "https://www.redtube.com", false);
		CreateBookmark(cat, "xHamster", "https://xhamster.com", false);
		CreateBookmark(cat, "Tube8", "https://www.tube8.com/", false);
		CreateBookmark(cat, "Xnxx", "https://www.xnxx.com/", false);
		CreateBookmark(cat, "Beeg", "https://beeg.com/", false);
		CreateBookmark(cat, "Flirt4Free", "https://www.flirt4free.com/", false);
		CreateBookmark(cat, "LiveJasmin", "https://www.livejasmin.com", false);
		CreateBookmark(cat, "Chaturbate", "https://chaturbate.com", false);
		CreateBookmark(cat, "Lush Stories", "https://www.lushstories.com/", false);
		CreateBookmark(cat, "Literotica", "https://literotica.com/", false);

		cat = CreateCategory("Live Places", false);
		CreateBookmark(cat, "Trevi Fountain - Rome", "https://www.skylinewebcams.com/en/webcam/italia/lazio/roma/fontana-di-trevi.html", false);
		CreateBookmark(cat, "Venice - Rialto Bridge", "https://www.skylinewebcams.com/en/webcam/italia/veneto/venezia/rialto-canal-grande.html", false);
		CreateBookmark(cat, "Fort Myers Beach - Florida", "https://www.skylinewebcams.com/en/webcam/united-states/florida/fort-myers-beach/fort-myers-beach.html", false);
		CreateBookmark(cat, "Colosseum - Rome", "https://www.skylinewebcams.com/en/webcam/italia/lazio/roma/colosseo.html", false);
		CreateBookmark(cat, "Statue of Liberty - New York", "https://www.skylinewebcams.com/en/webcam/united-states/new-york/new-york/statue-of-liberty.html", false);
		CreateBookmark(cat, "Lofoten Islands - Hamny", "https://www.skylinewebcams.com/en/webcam/norge/nordland/lofoten/hamnoy.html", false);
		CreateBookmark(cat, "The Acropolis of Athens", "https://www.skylinewebcams.com/en/webcam/ellada/atiki/athina/acropolis.html", false);
		CreateBookmark(cat, "The Pyramids of Giza and the Sphinx", "https://www.skylinewebcams.com/en/webcam/egypt/cairo/cairo/pyramids-giza-sphinx.html", false);

		cat = CreateCategory("Support", false);
		CreateBookmark(cat, "Fabio Den - Patreon", "https://www.patreon.com/fabioden", false);
		CreateBookmark(cat, "Fabio Den - Hub", "https://hub.virtamate.com/members/fabio.7254/#resources", false);
		CreateBookmark(cat, "Meshed VR - Patreon", "https://www.patreon.com/meshedvr", false);
		CreateBookmark(cat, "Official Wiki", "https://hub.virtamate.com/wiki/", false);
	}

	#endregion

	#region Serialize

	private string fullpath()
    {
		return _saveFolder + "bookmarks.json";
	}

	public void LoadData()
    {
		if (FileManagerSecure.FileExists(fullpath()) == false)
			DefaultData();
		else 
			ImportFileSelected(fullpath());
	}

	private void ImportFileSelected(string path)
	{
		if (string.IsNullOrEmpty(path)) return;

		try
		{
			JSONClass jc = JSON.Parse(SuperController.singleton.ReadFileIntoString(path)) as JSONClass;
			this.StartCoroutine(LoadExportedJSON(jc));
		}
		catch (Exception exc)
		{
			SuperController.LogError($"Bookmarks: Failed to import selection: {exc}");
		}
	}

	private IEnumerator LoadExportedJSON(JSONClass jc)
	{
		if (jc.HasKey("categoryCount") == false)
		{
			SuperController.LogMessage("Invalid Bookmark File!");
			yield return 0;
		}

		int count = int.Parse(jc["categoryCount"].Value);
		int counterb = 0;

		for (int j = 1; j <= count; j++)
		{
			string name = jc["categoryName" + j.ToString()].Value;
			int bookcount = int.Parse(jc["bookmarkCount" + j.ToString()].Value);

			// create new category
			Category cat = CreateCategory(name, false);

			for (int k = 1; k <= bookcount; k++)
			{
				counterb++;
				string title = jc["bookmarkTitle" + counterb.ToString()].Value;
				string url = jc["bookmarkURL" + counterb.ToString()].Value;

				// create new bookmark
				CreateBookmark(cat, title, url, false);
			}
		}
	}

	public void SaveData()
	{
		if (isDeleting == false)
        {
			ExportFileSelected(fullpath());
		}
	}

	private void ExportFileSelected(string path)
	{
		if (string.IsNullOrEmpty(path)) return;
		if (!path.ToLower().EndsWith($".{_saveExt}")) path += $".{_saveExt}";

		try
		{
			JSONClass jc = new JSONClass();
			// jc = SuperController.singleton.GetSaveJSON(null, true, false);
			FillExportJSON(jc);

			SuperController.singleton.SaveJSON(jc, path);
		}
		catch (Exception exc)
		{
			SuperController.LogError($"Bookmarks: Failed to export selection: {exc}");
		}
	}

	public void FillExportJSON(JSONClass jc)
	{
		jc.Add("categoryCount", cates.Count.ToString());

		int counterc = 0;
		int counterb = 0;

		foreach (Category cat in cates)
		{
			counterc++;
			jc.Add("categoryName" + counterc.ToString(), cat.name);
			jc.Add("bookmarkCount" + counterc.ToString(), cat.books.Count.ToString());

			foreach (Bookmark book in cat.books)
			{
				counterb++;
				jc.Add("bookmarkTitle" + counterb.ToString(), book.title);
				jc.Add("bookmarkURL" + counterb.ToString(), book.url);
			}
		}
	}

	#endregion

	#region Install

	private void Install()
	{
		string newDir = _saveFolder;

		if (FileManagerSecure.DirectoryExists(newDir) == false)
			FileManagerSecure.CreateDirectory(newDir);
	}

	#endregion

	#region Update

	public void Update()
	{

	}
	
	#endregion

} // End class

